/*
 * Each submatrix phi_i is stored by rows
 *
 * Each process operates on "nrows". However, processes allocate
 * room for two extra (nrows + 2) rows. This allows receiving data from
 * top and bottom neighbors in rows 0 and nrows+1 of the allocated block,
 * respectively, facilitating the application of the stencil.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>

#include "mpi.h"

#define MIN(a,b) (a) > (b) ? (b) : (a)
#define DEBUG 0
#define THRESHOLD (5e-3)

void init_phi( double * phi, int n, int my_first_row, int my_last_row );
void update( double *cur, double *next, int myrank, int nprocs, 
             int n, int my_first_row, int my_last_row ) ;
int converged( double *cur, double *next, int n, int my_first_row, int my_last_row );

int main( int argc, char *argv[] ) 
{
    int myrank, nprocs;
    int n, block_size, my_first_row, my_last_row;
    int niters, conv, all_conv;
    double *phi_cur, *phi_next, *tmp;
#if DEBUG
    double *full_phi;
#endif

    MPI_Init(&argc, &argv);

    MPI_Comm_rank( MPI_COMM_WORLD, &myrank );
    MPI_Comm_size( MPI_COMM_WORLD, &nprocs );

    if (argc != 2)
    {
        if ( myrank == 0 )
        {
            fprintf( stderr, "Usage: %s <n>\n", argv[0] );
            exit( -1 );
        }
    }

    n = atoi( argv[1] );

    if ( n < nprocs )
    {
        if ( myrank == 0 )
            fprintf(stderr, "[ERROR] nprocs > n. Come on, give me more work!\n" );
        MPI_Finalize();
        exit( -1 );
    }

    block_size = (n + nprocs-1) / nprocs;
    my_first_row = myrank * block_size;
    my_last_row = (myrank + 1) * block_size - 1;
    if ( n % nprocs != 0 )
        if ( myrank >= n % nprocs )
        {
            int offset = myrank - n % nprocs;
            my_first_row -= offset;
            my_last_row = MIN( my_first_row + (block_size) - 2, n-1 );
        }

    phi_cur  = (double *) malloc ( n * (block_size+2) * sizeof(double) );
    phi_next = (double *) malloc ( n * (block_size+2) * sizeof(double) );
#if DEBUG
    full_phi = (double *) malloc ( n * (nprocs*block_size) * sizeof(double) );
#endif
    
    init_phi(phi_cur, n, my_first_row, my_last_row);
    init_phi(phi_next, n, my_first_row, my_last_row);

    niters = 0;
    while( 1 )
    {
        niters++;
#if DEBUG
        if ( myrank == 0 )
            printf("Iteration %d\n", niters );

        // All blocks of rows gathered at root, 
        // which prints the full matrix
        MPI_Gather( &phi_cur[n], block_size*n, MPI_DOUBLE,
                    full_phi, block_size*n, MPI_DOUBLE,
                    0, MPI_COMM_WORLD );

        if ( myrank == 0 )
        {
            int row, col;
            for ( row = 0; row < n; row++ )
            {
                printf("[" );
                for ( col = 0; col < n; col++ )
                    printf(" %10.6f ", full_phi[row*n + col] );
                printf("]\n" );
            }
            sleep(1);
        }
#endif
        // Compute next (new) phi from current (old) phi
        update( phi_cur, phi_next, myrank, nprocs, n, my_first_row, my_last_row );
        // If converged, we are done
        conv = converged( phi_cur, phi_next, n, my_first_row, my_last_row );
        MPI_Allreduce( &conv, &all_conv, 1, MPI_INT, MPI_BAND, MPI_COMM_WORLD );
        if ( all_conv )
            break;
        // Otherwise, swap pointers and continue
        tmp = phi_cur;
        phi_cur = phi_next;
        phi_next = tmp;
    }

    free( phi_cur );
    free( phi_next );
#if DEBUG
    free( full_phi );
#endif

    if ( myrank == 0 )
        printf( "Converged after %d iterations\n", niters );

    MPI_Finalize();

    return 0;
}


void init_phi( double * phi, int n, int my_first_row, int my_last_row )
{
    int row, col;
    int nrows = my_last_row - my_first_row + 1;

    // Interior points initialized to 50 degrees
    for ( row = 1; row < nrows+1; row++ ) // row 1 to nrows (0 and nrows+1 are buffers for communication)
        for ( col = 1; col < n-1; col++ )
            phi[row*n + col] = 50.0;

    // If rank 0: top boundary fixed at 100 degrees
    if ( my_first_row == 0 )
    {
        row = 1;
        for ( col = 0; col < n; col++ )
            phi[row*n + col] = 100.0;
    }
    // Left, and right boundaries fixed at 100 degrees
    for ( row = 1; row < nrows+1; row++ )
    {
        phi[row*n +   0  ] = 100.0;
        phi[row*n + (n-1)] = 100.0;
    }
    // If rank nprocs - 1: bottom boundary fixed at 0 degrees
    if ( my_last_row == n-1 )
    {
        row = nrows;
        for ( col = 0; col < n; col++ )
            phi[row*n + col] = 0.0;
    }
}

void update( double *cur, double *next, int myrank, int nprocs, 
             int n, int my_first_row, int my_last_row )
{
    int row, col;
    int nrows = my_last_row - my_first_row + 1;

    // Even send first, then receive
    if ( myrank % 2 == 0 )
    {
        // Send to top
        if ( myrank != 0 )
            MPI_Ssend( &cur[n], n, MPI_DOUBLE, myrank-1, 0, MPI_COMM_WORLD );
        // Send to bottom
        if ( myrank != nprocs-1 )
            MPI_Ssend( &cur[nrows * n], n, MPI_DOUBLE, myrank+1, 1, MPI_COMM_WORLD );
        // Recv from top
        if ( myrank != 0 )
            MPI_Recv( &cur[0], n, MPI_DOUBLE, myrank-1, 1, MPI_COMM_WORLD, MPI_STATUS_IGNORE );
        // Recv from bottom
        if ( myrank != nprocs-1 )
            MPI_Recv( &cur[(nrows+1) * n], n, MPI_DOUBLE, myrank+1, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE );
    }
    // Odd recv first, then send
    else
    {
        // Recv from top
        if ( myrank != 0 )
            MPI_Recv( &cur[0], n, MPI_DOUBLE, myrank-1, 1, MPI_COMM_WORLD, MPI_STATUS_IGNORE );
        // Recv from bottom
        if ( myrank != nprocs-1 )
            MPI_Recv( &cur[(nrows+1) * n], n, MPI_DOUBLE, myrank+1, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE );
        // Send to top
        if ( myrank != 0 )
            MPI_Ssend( &cur[n], n, MPI_DOUBLE, myrank-1, 0, MPI_COMM_WORLD );
        // Send to bottom
        if ( myrank != nprocs-1 )
            MPI_Ssend( &cur[nrows * n], n, MPI_DOUBLE, myrank+1, 1, MPI_COMM_WORLD );
    }

    int frow = 1; // frow: first non-fixed row
    if ( myrank == 0 )
        frow += 1;
    int lrow = nrows; // lrow: last non-fixed row
    if ( myrank == nprocs-1 )
        lrow -= 1;

    for( row = frow; row <= lrow; row++ )
        for( col = 1; col < n-1; col++ )
            next[ row*n + col ] = 
                ( cur[ (row-1)*n +  col ] +
                  cur[  row   *n + (col-1) ] +
                  cur[  row   *n + (col+1) ] +
                  cur[ (row+1)*n +  col ] ) / 4;
}

int converged( double *cur, double *next, int n, int my_first_row, int my_last_row )
{
    int row, col;
    int nrows = my_last_row - my_first_row + 1;

    for( row = 1; row <= nrows; row++ )
        for( col = 1; col < n-1; col++ )
            if( fabs( next[ row*n + col ] - cur[ row*n + col ] ) > THRESHOLD )
                return 0;
    return 1;
}
