/*
 * Each submatrix phi_i is stored by columns
 *
 * Each process operates on "nrows". However, processes allocate
 * room for two extra (nrows + 2) rows. This allows receiving data from
 * top and bottom neighbors in rows 0 and nrows+1 of the allocated block,
 * respectively, facilitating the application of the stencil.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>

#include "mpi.h"

#define MIN(a,b) (a) > (b) ? (b) : (a)
#define DEBUG 0
#define THRESHOLD (5e-3)

void init_phi( double * phi, int n, int block_size, 
               int my_first_row, int my_last_row );
void update( double *cur, double *next, int myrank, int nprocs, 
             int n, int block_size, int my_first_row, int my_last_row, MPI_Datatype type );
int converged( double *cur, double *next, int n, int block_size, int my_first_row, int my_last_row );

int main( int argc, char *argv[] ) 
{
    int myrank, nprocs;
    int n, block_size, my_first_row, my_last_row;
    int niters, conv, all_conv;
    double *phi_cur, *phi_next, *tmp;
#if DEBUG
    double *full_phi;
#endif
    MPI_Datatype strided_vec;

    MPI_Init(&argc, &argv);

    MPI_Comm_rank( MPI_COMM_WORLD, &myrank );
    MPI_Comm_size( MPI_COMM_WORLD, &nprocs );

    if (argc != 2)
    {
        if ( myrank == 0 )
        {
            fprintf( stderr, "Usage: %s <n>\n", argv[0] );
            exit( -1 );
        }
    }

    n = atoi( argv[1] );

    if ( n < nprocs )
    {
        if ( myrank == 0 )
            fprintf(stderr, "[ERROR] nprocs > n. Come on, give me more work!\n" );
        MPI_Finalize();
        exit( -1 );
    }

    block_size = (n + nprocs-1) / nprocs;
    my_first_row = myrank * block_size;
    my_last_row = (myrank + 1) * block_size - 1;
    if ( n % nprocs != 0 )
        if ( myrank >= n % nprocs )
        {
            int offset = myrank - n % nprocs;
            my_first_row -= offset;
            my_last_row = MIN( my_first_row + (block_size) - 2, n-1 );
        }

    phi_cur  = (double *) malloc ( n * (block_size+2) * sizeof(double) );
    phi_next = (double *) malloc ( n * (block_size+2) * sizeof(double) );
#if DEBUG
    if ( myrank == 0 )
        full_phi = (double *) malloc ( n * (nprocs*block_size) * sizeof(double) );
#endif

    // New type to send/recv strided vectors (rows of a column-wise stored phi)
    MPI_Type_vector( n, 1, block_size+2, MPI_DOUBLE, &strided_vec );
    MPI_Type_commit( &strided_vec );
    
    init_phi( phi_cur, n, block_size, my_first_row, my_last_row );
    init_phi( phi_next, n, block_size, my_first_row, my_last_row );

    niters = 0;
    while( 1 )
    {
        niters++;
#if DEBUG
        if ( myrank == 0 )
            printf("Iteration %d\n", niters );

        // All blocks of rows gathered at root, 
        // which prints the full matrix
        int i, j;
        int displacements[nprocs];
        int recv_counts[nprocs];

        for ( i = 0; i < nprocs; i++ )
        {
            recv_counts[i] = n;
            displacements[i] = i*block_size*n;
        }

        for ( i = 0; i < block_size; i++ )
            MPI_Gatherv( &phi_cur[i+1], 1, strided_vec,
                         &full_phi[i*n], recv_counts, displacements, MPI_DOUBLE,
                         0, MPI_COMM_WORLD );
        // Now, full_phi is stored by rows
             
        if ( myrank == 0 )
        {
            int row, col;
            for ( row = 0; row < n; row++ )
            {
                printf("[" );
                for ( col = 0; col < n; col++ )
                    printf(" %10.6f ", full_phi[row*n + col] );
                printf("]\n" );
            }
            sleep(1);
        }
#endif
        // Compute next (new) phi from current (old) phi
        update( phi_cur, phi_next, myrank, nprocs, n, block_size, my_first_row, my_last_row, strided_vec );
        // If converged, we are done. All submatrices must converge at once
        conv = converged( phi_cur, phi_next, n, block_size, my_first_row, my_last_row );
        MPI_Allreduce( &conv, &all_conv, 1, MPI_INT, MPI_BAND, MPI_COMM_WORLD );
        if ( all_conv )
            break;
        // Otherwise, swap pointers and continue
        tmp = phi_cur;
        phi_cur = phi_next;
        phi_next = tmp;
    }

    free( phi_cur );
    free( phi_next );
#if DEBUG
    if ( myrank == 0 )
        free( full_phi );
#endif

    if ( myrank == 0 )
        printf( "Converged after %d iterations\n", niters );

    MPI_Type_free( &strided_vec );

    MPI_Finalize();

    return 0;
}

void init_phi( double * phi, int n, int block_size, int my_first_row, int my_last_row )
{
    int row, col;
    int ld = block_size + 2;
    int nrows = my_last_row - my_first_row + 1;

    // Interior points initialized to 50 degrees
    for ( col = 1; col < n-1; col++ )
        for ( row = 1; row < nrows+1; row++ ) // row 1 to nrows (0 and nrows+1 are buffers for communication)
            phi[col*ld + row] = 50.0;

    // If rank 0: top boundary fixed at 100 degrees
    if ( my_first_row == 0 )
    {
        row = 1;
        for ( col = 0; col < n; col++ )
            phi[col*ld + row] = 100.0;
    }
    // Left, and right boundaries fixed at 100 degrees
    for ( row = 1; row < nrows+1; row++ )
    {
        col = 0;
        phi[col*ld + row ] = 100.0;
        col = n-1;
        phi[col*ld + row ] = 100.0;
    }
    // If rank nprocs - 1: bottom boundary fixed at 0 degrees
    if ( my_last_row == n-1 )
    {
        row = nrows;
        for ( col = 0; col < n; col++ )
            phi[col*ld + row] = 0.0;
    }
}

void update( double *cur, double *next, int myrank, int nprocs, 
             int n, int block_size, int my_first_row, int my_last_row, MPI_Datatype strided_vec )
{
    int row, col;
    int ld = block_size + 2;
    int nrows = my_last_row - my_first_row + 1;

    MPI_Request req_send_top, 
                req_send_bot, 
                req_recv_top, 
                req_recv_bot;

    // Exchange with top neighbor
    if ( myrank != 0 )
    {
        MPI_Isend( &cur[1], 1, strided_vec, myrank-1, 0, MPI_COMM_WORLD, &req_send_top );
        MPI_Irecv( &cur[0], 1, strided_vec, myrank-1, 1, MPI_COMM_WORLD, &req_recv_top );
    }
    // Exchange with bottom neighbor
    if ( myrank != nprocs-1 )
    {
        MPI_Isend( &cur[nrows  ], 1, strided_vec, myrank+1, 1, MPI_COMM_WORLD, &req_send_bot );
        MPI_Irecv( &cur[nrows+1], 1, strided_vec, myrank+1, 0, MPI_COMM_WORLD, &req_recv_bot );
    }

    // Operate on the inner cells of my block
    for( col = 1; col < n-1; col++ )
        for( row = 2; row <= nrows-1; row++ )
            next[ col*ld + row ] = 
                ( cur[ (col-1)*ld +  row ] +
                  cur[  col   *ld + (row-1) ] +
                  cur[  col   *ld + (row+1) ] +
                  cur[ (col+1)*ld +  row ] ) / 4;

    // Ranks > 0 wait for row from top neighbor and operate on local first row
    if ( myrank != 0 )
    {
        MPI_Wait( &req_recv_top, MPI_STATUS_IGNORE );
        //  If last rank has one single row, it is constant.
        //  Do not modify it. 
        if ( !(myrank == nprocs-1 && nrows == 1) )
        {
            row = 1;
            for( col = 1; col < n-1; col++ )
                next[ col*ld + row ] = 
                    ( cur[ (col-1)*ld +  row ] +
                      cur[  col   *ld + (row-1) ] +
                      cur[  col   *ld + (row+1) ] +
                      cur[ (col+1)*ld +  row ] ) / 4;
        }
    }
    // Ranks < nprocs-1 wait for row from bottom neighbor and operate on local last row
    if ( myrank != nprocs-1 )
    {
        MPI_Wait( &req_recv_bot, MPI_STATUS_IGNORE );
        //  If first rank has one single row, it is constant.
        //  Do not modify it. 
        if ( !(myrank == 0 && nrows == 1) )
        {
            row = nrows;
            for( col = 1; col < n-1; col++ )
                next[ col*ld + row ] = 
                    ( cur[ (col-1)*ld +  row ] +
                      cur[  col   *ld + (row-1) ] +
                      cur[  col   *ld + (row+1) ] +
                      cur[ (col+1)*ld +  row ] ) / 4;
        }
    }

    // Cleanup. Wait for remaining requests
    if ( myrank != 0 )
        MPI_Wait( &req_send_top, MPI_STATUS_IGNORE );
    if ( myrank != nprocs-1 )
        MPI_Wait( &req_send_bot, MPI_STATUS_IGNORE );
}

int converged( double *cur, double *next, int n, int block_size, 
               int my_first_row, int my_last_row )
{
    int row, col;
    int ld = block_size + 2;
    int nrows = my_last_row - my_first_row + 1;

    for( col = 1; col < n-1; col++ )
        for( row = 1; row <= nrows; row++ )
            if( fabs( next[ col*ld + row ] - cur[ col*ld + row ] ) > THRESHOLD )
                return 0;
    return 1;
}
